from setup import flight_setup

import numpy as np
import openmdao.api as om

class Flight(om.Group):
    def setup(self):
        # mission profile parameters
        cfg = flight_setup.get()
        # trim components
        for i, scn in enumerate(['cruise', 'maneuver']):
            self.add_subsystem(f'trim_{scn}', Trim(wf=cfg['fixed_weight'], fi=cfg['init_fuel'], q=cfg['dynamic_pressure'][i], s=cfg['ref_area'], g=cfg['gravity']))
        # fuel volume component
        self.add_subsystem('volume_match', VolumeMatch(fr=cfg['res_fuel'], fi=cfg['init_fuel']))
        # fuel burn component
        self.add_subsystem('fuel_burn', FuelBurn(wf=cfg['fixed_weight'], r=cfg['range'], v=cfg['v_inf'], ct=cfg['sfc'], cd0=cfg['cd_0']))
        # fuel match component
        self.add_subsystem('fuel_match', FuelMatch(fi=cfg['init_fuel']))

class Trim(om.ExplicitComponent):
    """Vertical trim equation (mid-cruise): n = L / (Wland+Wfuel/2)
    """
    def initialize(self):
        self.options.declare('wf', desc='fixed aircraft weight, kg')
        self.options.declare('fi', desc='baseline fuel weight capacity, kg')
        self.options.declare('q', desc='dynamic pressure, Pa')
        self.options.declare('s', desc='reference area, m^2')
        self.options.declare('g', desc='gravity, m/s^2')
    def setup(self):
        self.add_input('cl', val=0., desc='lift coefficient')
        self.add_input('ww', val=0., units='kg', desc='half wing mass')
        self.add_input('fa', val=0., desc='fuel available for cruise normalized by baseline fuel capacity')
        self.add_output('n', val=0., desc='load factor')
        self.wf = self.options['wf']
        self.fi = self.options['fi']
        self.q = self.options['q']
        self.s = self.options['s']
        self.g = self.options['g']
        # Partials
        self.declare_partials(of='n', wrt=['cl', 'ww', 'fa'])
    def compute(self, inputs, outputs):
        outputs['n'] =  self.q * self.s / self.g * inputs['cl'] / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi)
    def compute_partials(self, inputs, partials):
        partials['n', 'cl'] = self.q * self.s / self.g / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi)
        partials['n', 'ww'] = -2 * self.q * self.s / self.g * inputs['cl'] / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi) / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi)
        partials['n', 'fa'] = -0.5 * self.fi * self.q * self.s / self.g * inputs['cl'] / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi) / (self.wf + 2 * inputs['ww'] + 0.5 * inputs['fa'] * self.fi)

class FuelBurn(om.ExplicitComponent):
    """Breguet formula to compute fuel burn during cruise: Wfuel = Wland * (e^(R*SFC*cd/cl)-1)
    """
    def initialize(self):
        self.options.declare('wf', desc='fixed aircraft weight, kg')
        self.options.declare('r', desc='target range, nm')
        self.options.declare('v', desc='true airspeed, kts')
        self.options.declare('ct', desc='specific fuel consumption, 1/h')
        self.options.declare('cd0', desc='aircraft zero-lift drag coefficient, -')
    def setup(self):
        # I/O
        self.add_input('ww', val=0., units='kg', desc='half wing mass')
        self.add_input('cl', val=0., desc='lift coefficient')
        self.add_input('cd', val=0., desc='drag coefficient')
        self.add_output('fb', val=0., units='kg', desc='fuel burn')
        self.wf = self.options['wf']
        self.r = self.options['r']
        self.v = self.options['v']
        self.ct = self.options['ct']
        self.cd0 = self.options['cd0']
        # Partials
        self.declare_partials(of='fb', wrt=['ww', 'cl', 'cd'])
    def __expo(self, cd, cl):
        return np.exp(self.r * self.ct / self.v * cd / cl)
    def compute(self, inputs, outputs):
        outputs['fb'] =  (self.wf + 2 * inputs['ww']) * (self.__expo(inputs['cd'] + self.cd0, inputs['cl']) - 1)
    def compute_partials(self, inputs, partials):
        cd = inputs['cd'] + self.cd0
        partials['fb', 'ww'] = 2 * (self.__expo(cd, inputs['cl']) - 1)
        partials['fb', 'cl'] = -(self.wf + 2 * inputs['ww']) * self.__expo(cd, inputs['cl']) * self.r * self.ct / self.v * cd / inputs['cl'] / inputs['cl']
        partials['fb', 'cd'] = (self.wf + 2 * inputs['ww']) * self.__expo(cd, inputs['cl']) * self.r * self.ct / self.v / inputs['cl']

class FuelMatch(om.ExplicitComponent):
    """Fuel matching: ratio = Wfuel_avail / Wfuel_burn
    """
    def initialize(self):
        self.options.declare('fi', desc='baseline fuel weight capacity, kg')
    def setup(self):
        self.add_input('fa', val=0., desc='fuel available for cruise normalized by baseline fuel capacity')
        self.add_input('fb', val=0., units='kg', desc='fuel burn')
        self.add_output('fm', val=0., desc='fuel match ratio')
        self.fi = self.options['fi']
        # Partials
        self.declare_partials(of='fm', wrt=['fa', 'fb'])
    def compute(self, inputs, outputs):
        outputs['fm'] =  inputs['fa'] * self.fi / inputs['fb']
    def compute_partials(self, inputs, partials):
        partials['fm', 'fa'] = self.fi / inputs['fb']
        partials['fm', 'fb'] = -inputs['fa'] * self.fi / (inputs['fb'] * inputs['fb'])

class VolumeMatch(om.ExplicitComponent):
    """Volume matching: ratio = Vwing_avail / Vwing_init * Wfuel_init / (Wfuel_avail + Wreserve)
    """
    def initialize(self):
        self.options.declare('fr', desc='reserve fuel weight, kg')
        self.options.declare('fi', desc='baseline fuel weight capacity, kg')
    def setup(self):
        self.add_input('vr', val=0., desc='ratio of current volume to intial volume')
        self.add_input('fa', val=0., desc='fuel available for cruise normalized by baseline fuel capacity')
        self.add_output('vc', val=0., desc='volume constraint')
        self.fr = self.options['fr']
        self.fi = self.options['fi']
        # Partials
        self.declare_partials(of='vc', wrt=['vr', 'fa'])
    def compute(self, inputs, outputs):
        outputs['vc'] = inputs['vr'] / (inputs['fa'] + self.fr / self.fi)
    def compute_partials(self, inputs, partials):
        partials['vc', 'vr'] = 1 / (inputs['fa'] + self.fr / self.fi)
        partials['vc', 'fa'] = -inputs['vr'] / (inputs['fa'] + self.fr / self.fi) / (inputs['fa'] + self.fr / self.fi)
