from setup.struct_dv_setup import struct_comps
from static import Static
from struct_dv import StructDvMapper, SmoothnessEvaluatorGrid
from flight import Flight

import numpy as np
import openmdao.api as om

class Top(om.Group):
    def setup(self):
        # MPhys
        # create ivc to keep the top level DVs
        self.add_subsystem('dvs', om.IndepVarComp(), promotes=['*'])
        # create mapper for structural DVs
        self.add_subsystem('struct_mapper', StructDvMapper(), promotes=['*'])
        # create multipoint group
        self.add_subsystem('static', Static())
        # structural DV smoothness component
        self.add_subsystem('spar_smoothness', SmoothnessEvaluatorGrid(columns=struct_comps['sizes']['spar'], rows=1))
        self.add_subsystem('skin_smoothness', SmoothnessEvaluatorGrid(columns=struct_comps['sizes']['skin'], rows=1))
        # mission profile parameters
        self.add_subsystem('flight', Flight())

    def configure(self):
        # DV - fuel
        self.dvs.add_output('av_fuel', val=110.0/135.0, desc='fuel available for cruise normalized by baseline fuel capacity')
        """self.add_design_var('av_fuel', lower=0.5, upper=1.5, scaler=1.0)"""
        # DV - aoa
        aoas = [3.0, 4.0]
        for i, scn in enumerate(['cruise', 'maneuver']):
            self.dvs.add_output(f'aoa_{scn}', val=aoas[i], units='deg')
            self.connect(f'aoa_{scn}', f'static.{scn}.aoa')
            self.add_design_var(f'aoa_{scn}', lower=0.0, upper=6.0, scaler=1.0, units='deg')
        # DV - struct
        for name, value in struct_comps['values'].items():
            self.dvs.add_output(name, val=value)
        self.connect('spar','spar_smoothness.thickness')
        self.connect('skin','skin_smoothness.thickness')
        for scn in ['cruise', 'maneuver']:
            self.connect('dv_struct', f'static.{scn}.dv_struct')
        """for name, value in struct_comps['values'].items():
            self.dvs.add_design_var(name, lower=0.001, upper=0.2, scaler=np.reciprocal(value))"""
        # DV - shape (global)
        self.dvs.add_output('twist', val=np.array([0] * self.static.geo_ntwist))
        for scn in ['cruise', 'maneuver']:
          self.connect('twist', f'static.{scn}.twist')
        """self.add_design_var('twist', lower=-5.0, upper=5.0, scaler=1.0)"""
        # DV - shape (local)
        self.dvs.add_output('shape', val=np.array([0] * self.static.geo_nshape))
        for scn in ['cruise', 'maneuver']:
            self.connect('shape', f'static.{scn}.shape')
        """self.add_design_var('shape', lower=-0.1, upper=0.1, scaler=10.0)"""
        # CON - trim
        for scn in ['cruise', 'maneuver']:
            self.connect(f'static.{scn}.cl', f'flight.trim_{scn}.cl')
            self.connect(f'static.cruise.mass', f'flight.trim_{scn}.ww')
            self.connect('av_fuel', f'flight.trim_{scn}.fa')
        self.add_constraint('flight.trim_cruise.n', equals=1.0, scaler=1.0)
        self.add_constraint('flight.trim_maneuver.n', equals=2.5, scaler=0.4)
        # CON - fuel matching
        self.connect('av_fuel', 'flight.fuel_match.fa')
        self.connect('flight.fuel_burn.fb', 'flight.fuel_match.fb')
        self.add_constraint('flight.fuel_match.fm', lower=1.0, upper=1.05, scaler=1.0)
        # CON - struct
        self.add_constraint(f'static.maneuver.ks_fail', upper=1.0, scaler=1.0)
        """self.add_constraint('spar_smoothness.diff', ref=5e-3, upper=0.0, linear=True)
        self.add_constraint('skin_smoothness.diff', ref=5e-3, upper=0.0, linear=True)"""
        # CON - shape
        """self.add_constraint('static.cruise.thckle_con', lower=1.0, scaler=1.0)
        self.add_constraint('static.cruise.thckte_con', lower=1.0, scaler=1.0)
        self.add_constraint('static.cruise.thcksp_con', lower=0.8, scaler=1.0)
        self.add_constraint('static.cruise.le_con', equals=0.0, scaler=1.0, linear=True)
        self.add_constraint('static.cruise.te_con', equals=0.0, scaler=1.0, linear=True)"""
        # CON - shape (fuel tank volume)
        self.connect('static.cruise.vr', 'flight.volume_match.vr', src_indices=[0])
        self.connect('av_fuel', 'flight.volume_match.fa')
        """self.add_constraint('flight.volume_match.vc', lower=1.0, scaler=1.0)"""
        # OBJ
        self.connect('static.cruise.mass', 'flight.fuel_burn.ww')
        self.connect('static.cruise.cl', 'flight.fuel_burn.cl')
        self.connect('static.cruise.cd', 'flight.fuel_burn.cd')
        self.add_objective('flight.fuel_burn.fb', scaler=1e-5)

def main():
    from utils import change_workdir, print_time
    import time
    from mpi4py import MPI
    # Change to working directory
    change_workdir()
    # Setup
    prob = om.Problem(model=Top(), driver=om.ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-4, disp=True, maxiter=3, debug_print=['desvars', 'nl_cons', 'ln_cons']))
    if MPI.COMM_WORLD.size == 1:
        prob.model.recording_options['includes'] = ['av_fuel', 'aoa_cruise', 'aoa_maneuver', 'twist', 'shape', 'ribs', 'le_spar', 'te_spar','up_skin', 'lo_skin', 'up_stringer', 'lo_stringer', 'static.cruise.cl', 'static.cruise.cd', 'static.maneuver.cl', 'static.maneuver.cd', 'static.cruise.ks_fail', 'static.maneuver.ks_fail', 'static.cruise.mass', 'flight.trim_cruise.n', 'flight.trim_maneuver.n', 'flight.volume_match.vc', 'flight.fuel_burn.fb', 'flight.fuel_match.fm']
        prob.model.add_recorder(om.SqliteRecorder('cases.sql'))
    prob.setup(mode='rev')
    om.n2(prob, show_browser=False, outfile='mphys_dart_tacs.html')
    # Run
    tic = time.perf_counter()
    #prob.run_model()
    #prob.check_totals(of='fuel.fb', wrt=['aoa','ribs','twist'], compact_print=False, form='central', show_progress=True, driver_scaling=False)
    prob.run_driver()
    toc = time.perf_counter()
    # Get variables
    aoas = []
    cls = []
    cds = []
    ns = []
    kss = []
    dzs = []
    for scn in ['cruise', 'maneuver']:
        aoas.append(prob.get_val(f'aoa_{scn}', get_remote=True)[0])
        cls.append(prob.get_val(f'static.{scn}.cl', get_remote=True)[0])
        cds.append(prob.get_val(f'static.{scn}.cd', get_remote=True)[0])
        ns.append(prob.get_val(f'flight.trim_{scn}.n', get_remote=True)[0])
        kss.append(prob.get_val(f'static.{scn}.ks_fail', get_remote=True)[0])
        dzs.append(max(prob.get_val(f'static.{scn}.u_struct', get_remote=True)[2::3]))
    # Display
    if MPI.COMM_WORLD.rank == 0:
        for i, scn in enumerate(['cruise', 'maneuver']):
            print(f'{scn}:'.capitalize())
            print('  aoa =', aoas[i], 'deg')
            print('  cl =', cls[i])
            print('  cd =', cds[i])
            print('  load factor =', ns[i])
            print('  KSFailure =', kss[i])
            print('  z-deflection =', dzs[i], 'm')
        print('Geo:')
        print('  mass =', prob['static.cruise.mass'][0], 'kg')
        print('  twist =', prob['twist'])
        print('  volume matching =', prob['flight.volume_match.vc'][0])
        print('Perfo:')
        print('  fuel matching =', prob['flight.fuel_match.fm'][0])
        print('  fuel burn =', prob['flight.fuel_burn.fb'][0], 'kg')
        print_time(toc-tic)
    # Save
    if MPI.COMM_WORLD.size == 1:
        cases = om.CaseReader('cases.sql').get_cases()
        matrix = np.zeros((len(cases), 13))
        for i, case in enumerate(cases):
            matrix[i,0] = i
            matrix[i,1] = case.get_design_vars(scaled=False)['aoa_cruise'][0]
            matrix[i,2] = case.get_design_vars(scaled=False)['aoa_maneuver'][0]
            matrix[i,3] = case.get_val('static.cruise.cl')[0]
            matrix[i,4] = case.get_val('static.maneuver.cl')[0]
            matrix[i,5] = case.get_val('static.cruise.cd')[0]
            matrix[i,6] = case.get_val('static.cruise.mass')[0]
            matrix[i,7] = case.get_constraints(scaled=False)['flight.trim_cruise.n'][0]
            matrix[i,8] = case.get_constraints(scaled=False)['flight.trim_maneuver.n'][0]
            matrix[i,9] = case.get_constraints(scaled=False)['static.maneuver.ks_fail'][0]
            #matrix[i,10] = case.get_constraints(scaled=False)['flight.volume_match.vc'][0]
            matrix[i,11] = case.get_constraints(scaled=False)['flight.fuel_match.fm'][0]
            matrix[i,12] = case.get_objectives(scaled=False)['flight.fuel_burn.fb'][0]
        np.savetxt('history.dat', matrix, header='it, aoa0, aoa1, cl0, cl1, cd0, ww, n0, n1, ks1, vol, fm, wf')

if __name__ == '__main__':
    main()
