from setup.struct_dv_setup import struct_comps 
import openmdao.api as om

class StructDvMapper(om.ExplicitComponent):
    """OpenMDAO component to map the user-defined DV to those defined in NASTRAN/TACS
    """
    def setup(self):
        for name, size in struct_comps['sizes'].items():
            self.add_input(name, shape=size)
        self.add_output('dv_struct', shape=sum(struct_comps['sizes'].values()))

    def compute(self, inputs, outputs):
        for name, size in struct_comps['sizes'].items():
            for j in range(size):
                outputs['dv_struct'][struct_comps['indices'][name][j]] = inputs[name][j]

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        if mode == 'fwd':
            raise NotImplementedError('StructDvMapper - forward mode not implemented!\n')
        if mode == 'rev':
            if 'dv_struct' in d_outputs:
                for name, size in struct_comps['sizes'].items():
                    for j in range(size):
                        d_inputs[name][j] += d_outputs['dv_struct'][struct_comps['indices'][name][j]]

class SmoothnessEvaluatorGrid(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('delta',default=0.001) # [m]
        self.options.declare('rows',default=1)
        self.options.declare('columns',default=1)

    def setup(self):
        self.rows = self.options['rows']
        self.columns = self.options['columns']

        self.ndiffs = 2 * (self.columns-1) * self.rows + 2 * self.columns * (self.rows-1)

        # row major thicknesses
        self.add_input('thickness',shape=self.rows*self.columns)
        self.add_output('diff',shape=self.ndiffs)

    def compute(self,inputs,outputs):
        delta = self.options['delta']

        j = 0
        for row in range(self.rows):
            for col in range(self.columns-1):
                i = self.columns * row + col
                outputs['diff'][j]   = inputs['thickness'][i]   - inputs['thickness'][i+1] - delta
                outputs['diff'][j+1] = inputs['thickness'][i+1] - inputs['thickness'][i]   - delta
                j += 2

        for row in range(self.rows-1):
            for col in range(self.columns):
                i  = self.columns * row + col
                i2 = self.columns * (row+1) + col
                outputs['diff'][j]   = inputs['thickness'][i]  - inputs['thickness'][i2] - delta
                outputs['diff'][j+1] = inputs['thickness'][i2] - inputs['thickness'][i]  - delta
                j += 2

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        if mode == 'fwd':
            raise NotImplementedError('SmoothnessEvaluatorGrid - forward mode not implemented!\n')
        if mode == 'rev':
            if 'diff' in d_outputs:
                if 'thickness' in d_inputs:
                    j = 0
                    for row in range(self.rows):
                        for col in range(self.columns-1):
                            i  = self.columns * row + col
                            d_inputs['thickness'][i]   += d_outputs['diff'][j]
                            d_inputs['thickness'][i+1] -= d_outputs['diff'][j]
                            d_inputs['thickness'][i+1] += d_outputs['diff'][j+1]
                            d_inputs['thickness'][i]   -= d_outputs['diff'][j+1]
                            j += 2

                    for row in range(self.rows-1):
                        for col in range(self.columns):
                            i  = self.columns * row + col
                            i2 = self.columns * (row+1) + col
                            d_inputs['thickness'][i]  += d_outputs['diff'][j]
                            d_inputs['thickness'][i2] -= d_outputs['diff'][j]
                            d_inputs['thickness'][i2] += d_outputs['diff'][j+1]
                            d_inputs['thickness'][i]  -= d_outputs['diff'][j+1]
                            j += 2

if __name__ == '__main__':
    from openmdao.api import Problem
    prob = Problem()
    prob.model.add_subsystem('StructDvDistributor',StructDvMapper())
    prob.model.add_subsystem('smoothness_grid',SmoothnessEvaluatorGrid(rows=3,columns=4))

    prob.setup(force_alloc_complex=True)
    prob.check_partials(method='cs',compact_print=True)
