from setup import pygeo_setup, dart_setup
from utils import change_workdir, parse_args, print_time

from geometry import PyGeoBuilder
from dartflo.dart.api.mphys import DartBuilder

from mphys import Multipoint, MPhysVariables
from mphys.scenarios.aerodynamic import ScenarioAerodynamic
import openmdao.api as om

class Top(Multipoint):
    def setup(self):
        # pyGeo
        geom = PyGeoBuilder(pygeo_setup.get())
        geom.initialize(self.comm)
        # Dart
        args = parse_args()
        dart = DartBuilder(dart_setup.get(args.k, args.v), scenario='aerodynamic', task='optimization', saveAll=False, raiseError=False)
        dart.initialize(self.comm)

        # MPhys
        # create ivc to keep the top level DVs
        self.add_subsystem('dvs', om.IndepVarComp(), promotes=['*'])
        # create the mesh
        self.add_subsystem('mesh', dart.get_mesh_coordinate_subsystem())
        # create the geometry
        self.add_subsystem('geometry', geom.get_mesh_coordinate_subsystem())
        # create the scenario
        self.mphys_add_scenario('cruise', ScenarioAerodynamic(aero_builder=dart))
        # connect
        self.connect(f'mesh.{MPhysVariables.Aerodynamics.Surface.Mesh.COORDINATES}', f'geometry.{MPhysVariables.Aerodynamics.Surface.Geometry.COORDINATES_INPUT}')
        self.connect(f'geometry.{MPhysVariables.Aerodynamics.Surface.Geometry.COORDINATES_OUTPUT}', f'cruise.{MPhysVariables.Aerodynamics.Surface.COORDINATES}')

    def configure(self):
        import numpy as np
        # DV - aoa
        self.dvs.add_output('aoa', val=3.06, units='deg')
        self.connect('aoa', 'cruise.aoa')
        self.add_design_var('aoa', lower=0.0, upper=5.0, scaler=0.1, units='deg')
        # DV - shape
        self.geometry.addDisciplinePoints('aero', self.mesh.get_val(MPhysVariables.Aerodynamics.Surface.Mesh.COORDINATES))
        sizes = self.geometry.getSizes()
        self.dvs.add_output('twist', val=np.zeros(sizes['twist']))
        self.dvs.add_output('shape', val=np.zeros(sizes['shape']))
        self.connect('twist', 'geometry.twist')
        self.connect('shape', 'geometry.shape')
        self.add_design_var('twist', lower=-5.0, upper=5.0, scaler=0.1)
        self.add_design_var('shape', lower=-0.01, upper=0.01, scaler=10.0)
        # CON - shape
        self.geometry.setConstrainedSurface(self.mesh.get_triangulated_surface())
        self.add_constraint('geometry.thckle_con', lower=1.0, scaler=1.0)
        self.add_constraint('geometry.thckte_con', lower=1.0, scaler=1.0)
        self.add_constraint('geometry.vol_con', lower=1.0, scaler=1.0)
        self.add_constraint('geometry.le_con', equals=0.0, scaler=1.0, linear=True)
        self.add_constraint('geometry.te_con', equals=0.0, scaler=1.0, linear=True)
        # CON - loads
        self.add_constraint('cruise.cl', equals=0.30, scaler=1.0)
        # OBJ - loads
        self.add_objective('cruise.cd', scaler=100.0)

def main():
    import time
    import numpy as np
    # Change to working directory
    change_workdir()
    # Setup
    prob = om.Problem(model=Top(), driver=om.ScipyOptimizeDriver(optimizer='SLSQP', tol=1e-3, disp=True))
    prob.driver.recording_options['includes'] = ['aoa','twist','shape','cruise.cl','cruise.cd']
    prob.driver.add_recorder(om.SqliteRecorder('cases.sql'))
    prob.setup(mode='rev')
    om.n2(prob, show_browser=False, outfile='mphys_dart.html')
    # Run
    tic = time.perf_counter()
    #prob.run_model()
    prob.run_driver()
    toc = time.perf_counter()
    # Printout
    print('Scenario:')
    print('aoa =', prob['aoa'][0],'deg')
    print('cl =', prob['cruise.cl'][0])
    print('cd =', prob['cruise.cd'][0])
    print('twist =', prob['twist'])
    print_time(toc-tic)
    # Save
    cases = om.CaseReader('cases.sql').get_cases()
    matrix = np.zeros((len(cases), 4))
    for i, case in enumerate(cases):
        matrix[i,0] = i
        matrix[i,1] = case.get_design_vars(scaled=False)['aoa'][0]
        matrix[i,2] = case.get_constraints(scaled=False)['cruise.cl'][0]
        matrix[i,3] = case.get_objectives(scaled=False)['cruise.cd'][0]
    np.savetxt('history.dat', matrix, header='it, aoa, cl, cd')

if __name__ == '__main__':
    main()
