## MPHYS interface for pyGeo
# Adrien Crovato

import numpy as np
from mpi4py import MPI
import openmdao.api as om
from omflut.builder import Builder as OMFlutBuilder
from mphys import Builder as MphysBuilder, MPhysVariables

# Define short aliases for convenience
XAERO_IN = MPhysVariables.Aerodynamics.Surface.Geometry.COORDINATES_INPUT
XAERO_OUT = MPhysVariables.Aerodynamics.Surface.Geometry.COORDINATES_OUTPUT
XSTRUCT_IN = MPhysVariables.Structures.Geometry.COORDINATES_INPUT
XSTRUCT_OUT = MPhysVariables.Structures.Geometry.COORDINATES_OUTPUT

# Geometry parametrization and constraints
class PyGeoGeom(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('dvGeo', desc='geometry parametrization', recordable=False)
        self.options.declare('dvCon', desc='constraints', recordable=False)
        self.options.declare('config', desc='configuration', recordable=False)
        self.options.declare('use_old_io', default=False, desc='use old in/out names', recordable=False)

    def setup(self):
        self.DVGeo = self.options['dvGeo']
        self.DVCon = self.options['dvCon']
        self.config = self.options['config']
        self.omPtSetList = []
        # Setup geometry, constraints and their I/O
        # add IO for surface meshes, the DVGeo object must be configured later by calling addDisciplinePoints
        if self.options['use_old_io']:
            for disc in self.config['disciplines']:
                self.add_input(f'x_{disc}_in', distributed=True, shape_by_conn=True)
                self.add_output(f'x_{disc}0', distributed=True, copy_shape=f'x_{disc}_in', tags=['mphys_coordinates'])
        else:
            if 'aero' in self.config['disciplines']:
                self.add_input(XAERO_IN, distributed=True, shape_by_conn=True, tags=['mphys_coordinates'])
                self.add_output(XAERO_OUT, distributed=True, copy_shape=XAERO_IN, tags=['mphys_coordinates'])
            if 'struct' in self.config['disciplines']:
                self.add_input(XSTRUCT_IN, distributed=True, shape_by_conn=True, tags=['mphys_coordinates'])
                self.add_output(XSTRUCT_OUT, distributed=True, copy_shape=XSTRUCT_IN, tags=['mphys_coordinates'])
        # add a reference axis
        if 'refAxis' in self.config:
            for name, prm in self.config['refAxis'].items():
                self.DVGeo.addRefAxis(name, xFraction=prm['xFraction'], alignIndex=prm['alignIndex'])
        # add global DV
        if 'globalDV' in self.config:
            for name, prm in self.config['globalDV'].items():
                self.DVGeo.addGlobalDV(name, prm['value'], prm['func'])
                self.add_input(name, distributed=False, shape=len(prm['value']), tags=['mphys_input'])
        # add local DV
        if 'localDV' in self.config:
            for name, prm in self.config['localDV'].items():
                nVal = self.DVGeo.addLocalDV(name, axis=prm['axis'])
                self.add_input(name, distributed=False, shape=nVal, tags=['mphys_input'])
        # add leading edge radius constraint
        if 'leRadiusCon' in self.config:
            for name, prm in self.config['leRadiusCon'].items():
                if self.comm.rank == 0:
                    self.add_output(name, distributed=False, val=np.ones(prm['nSpan']), tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        # add thickness constraints
        if 'thckCon1D' in self.config:
            for name, prm in self.config['thckCon1D'].items():
                if self.comm.rank == 0:
                    self.add_output(name, distributed=False, val=np.ones(prm['nCon']), tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        if 'thckCon2D' in self.config:
            for name, prm in self.config['thckCon2D'].items():
                if self.comm.rank == 0:
                    self.add_output(name, distributed=False, val=np.ones(prm['nSpan'] * prm['nChord']), tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        # add volume constraints
        if 'volCon' in self.config:
            for name, prm in self.config['volCon'].items():
                if self.comm.rank == 0:
                    self.add_output(name, distributed=False, val=1.0, tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        # add LE/TE constraints
        if 'leteCon' in self.config:
            for name, prm in self.config['leteCon'].items():
                if self.comm.rank == 0:
                    nCon = self.__getLeTeConSize(prm['volID'], prm['faceID'])
                    self.add_output(name, distributed=False, val=np.zeros(nCon), tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        # add linear constraints on local dv
        if 'linConShape' in self.config:
            if len(self.config['linConShape']) > 1:
                raise RuntimeError('PyGeoGeom - Linear shape constraints must be unique!\n')
            for name, prm in self.config['linConShape'].items():
                if self.comm.rank == 0:
                    nCon = len(prm['indSetA'])
                    self.add_output(name, distributed=False, val=np.zeros(nCon), tags=['mphys_result'])
                else:
                    self.add_output(name, distributed=True, shape=0, tags=['mphys_result']) # TODO distributed
        # add VSP variables
        if 'vspVar' in self.config:
            import warnings
            warnings.warn('PyGeoGeom - VSP interface not tested!\n', RuntimeWarning) # TODO NOT TESTED
            for prm in self.config['vspVar']:
                self.DVGeo.addVariable(prm['component'], prm['group'], prm['parm'], **prm['kwargs'])
                name = '{component}:{group}:{parm}'.format(**prm)
                val = self.DVGeo.DVs[name].value.copy()
                self.add_input(name, distributed=False, shape=1, val=val, tags=['mphys_input'])

    def addDisciplinePoints(self, discipline, points, **kwargs):
        """Add the points of the discipline to be embedded in the parametrization
        """
        if self.options['use_old_io']:
            self.DVGeo.addPointSet(points.reshape(len(points) // 3, 3), f'x_{discipline}0', **kwargs)
            self.omPtSetList.append(f'x_{discipline}0')
        else:
            if discipline == 'aero':
                self.DVGeo.addPointSet(points.reshape(len(points) // 3, 3), XAERO_OUT, **kwargs)
                self.omPtSetList.append(XAERO_OUT)
            elif discipline == 'struct':
                self.DVGeo.addPointSet(points.reshape(len(points) // 3, 3), XSTRUCT_OUT, **kwargs)
                self.omPtSetList.append(XSTRUCT_OUT)

    def setConstrainedSurface(self, surface):
        """Set (a triangulation of) the surface to be constrained and add the constraints to DVCon
        """
        # Set the surface
        self.DVCon.setSurface(surface)
        # Add the constraints (can only be done after the surface has been set)
        if 'leRadiusCon' in self.config:
            for name, prm in self.config['leRadiusCon'].items():
                self.DVCon.addLERadiusConstraints(prm['leList'], prm['nSpan'], prm['axis'], prm['chordDir'], name=name)
        if 'thckCon1D' in self.config:
            for name, prm in self.config['thckCon1D'].items():
                self.DVCon.addThicknessConstraints1D(prm['ptList'], prm['nCon'], prm['axis'], name=name)
        if 'thckCon2D' in self.config:
            for name, prm in self.config['thckCon2D'].items():
                self.DVCon.addThicknessConstraints2D(prm['leList'], prm['teList'], prm['nSpan'], prm['nChord'], name=name)
        if 'volCon' in self.config:
            for name, prm in self.config['volCon'].items():
                self.DVCon.addVolumeConstraint(prm['leList'], prm['teList'], prm['nSpan'], prm['nChord'], name=name)
        if 'leteCon' in self.config:
            for name, prm in self.config['leteCon'].items():
                self.DVCon.addLeTeConstraints(prm['volID'], prm['faceID'], name=name)
        if 'linConShape' in self.config:
            for name, prm in self.config['linConShape'].items():
                self.DVCon.addLinearConstraintsShape(prm['indSetA'], prm['indSetB'], prm['factorA'], prm['factorB'], name=name)

    def getSizes(self):
        """Return the size of the design variables
        """
        dvs = self.DVGeo.getValues()
        for k, v in dvs.items():
            dvs[k] = len(v)
        return dvs

    def __getLeTeConSize(self, volID, faceID):
        """Return the number of LE/TE constraint.
           Borrowed from pygeo.DVConstraints.addLeTeConstraints
        """
        # Get the number of local DV per axis
        shp = self.DVGeo.getLocalIndex(volID).shape
        # Find the face
        if 'i' in faceID:
            ks = [1,2]
        elif 'j' in faceID:
            ks = [0,2]
        elif 'k' in faceID:
            ks = [0,1]
        else:
            raise RuntimeError('PyGeoGeom - wrong faceID provided!\n')
        # Find the axis
        if shp[ks[0]] == 2 and shp[ks[1]] > 2:
            return shp[ks[1]]
        elif shp[ks[0]] > 2 and shp[ks[1]] == 2:
            return shp[ks[0]]
        else:
            raise RuntimeError('PyGeoGeom - only one of the FFD dimensions should be of size 2!\n')

    def compute(self, inputs, outputs):
        # inputs are the geometric design variables
        self.DVGeo.setDesignVars(inputs)

        # ouputs are the coordinates of the pointsets we have
        for ptName in self.DVGeo.points:
            if ptName in self.omPtSetList:
                # update this pointset and write it as output
                outputs[ptName] = self.DVGeo.update(ptName).flatten()

        # compute the DVCon constraint values
        constraintfunc = dict()
        self.DVCon.evalFunctions(constraintfunc, includeLinear=True)
        comm = self.comm
        if comm.rank == 0:
            for constraintname in constraintfunc:
                outputs[constraintname] = constraintfunc[constraintname]

        # we ran a compute so the inputs changed. update the dvcon jac
        # next time the jacvec product routine is called
        self.update_jac = True

        # write deformed FFD
        self.DVGeo.writeTecplot('ffd_def.dat')
        self.DVCon.writeTecplot('ffd_con.dat')

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):
        # only do the computations when we have more than zero entries in d_inputs in the reverse mode
        ni = len(list(d_inputs.keys()))

        if mode == "rev" and ni > 0:

            # this flag will be set to True after every compute call.
            # if it is true, we assume the design has changed so we re-run the sensitivity update
            # there can be hundreds of calls to this routine due to thickness constraints,
            # as a result, we only run the actual sensitivity comp once and save the jacobians
            # this might be better suited with the matrix-based API
            if self.update_jac:
                self.constraintfuncsens = dict()
                self.DVCon.evalFunctionsSens(self.constraintfuncsens, includeLinear=True)
                # set the flag to False so we dont run the update again if this is called w/o a compute in between
                self.update_jac = False

            for constraintname in self.constraintfuncsens:
                for dvname in self.constraintfuncsens[constraintname]:
                    if dvname in d_inputs:
                        dcdx = self.constraintfuncsens[constraintname][dvname]
                        if self.comm.rank == 0:
                            dout = d_outputs[constraintname]
                            jvtmp = np.dot(np.transpose(dcdx), dout)
                        else:
                            jvtmp = 0.0
                        d_inputs[dvname] += jvtmp
                        # OM does the reduction itself
                        # d_inputs[dvname] += self.comm.reduce(jvtmp, op=MPI.SUM, root=0)

            for ptSetName in self.DVGeo.ptSetNames:
                if ptSetName in self.omPtSetList:
                    dout = d_outputs[ptSetName].reshape(len(d_outputs[ptSetName]) // 3, 3)

                    # only do the calc. if d_output is not zero on ANY proc
                    local_all_zeros = np.all(dout == 0)
                    global_all_zeros = np.zeros(1, dtype=bool)
                    # we need to communicate for this check otherwise we may hang
                    self.comm.Allreduce([local_all_zeros, MPI.BOOL], [global_all_zeros, MPI.BOOL], MPI.LAND)

                    # global_all_zeros is a numpy array of size 1
                    if not global_all_zeros[0]:

                        # TODO totalSensitivityTransProd is broken. does not work with zero surface nodes on a proc
                        # xdot = self.DVGeo.totalSensitivityTransProd(dout, ptSetName)
                        xdot = self.DVGeo.totalSensitivity(dout, ptSetName)

                        # loop over dvs and accumulate
                        xdotg = {}
                        for k in xdot:
                            # check if this dv is present
                            if k in d_inputs:
                                # do the allreduce
                                # TODO reove the allreduce when this is fixed in openmdao
                                # reduce the result ourselves for now. ideally, openmdao will do the reduction itself when this is fixed. this is because the bcast is also done by openmdao (pyoptsparse, but regardless, it is not done here, so reduce should also not be done here)
                                xdotg[k] = self.comm.allreduce(xdot[k], op=MPI.SUM)

                                # accumulate in the dict
                                # TODO
                                # because we only do one point set at a time, we always want the 0th
                                # entry of this array since dvgeo always behaves like we are passing
                                # in multiple objective seeds with totalSensitivity. we can remove the [0]
                                # once we move back to totalSensitivityTransProd
                                d_inputs[k] += xdotg[k][0]


# Builder
class PyGeoBuilder(MphysBuilder):
    """pyGeo builder for MPHYS

    Attributes
    ----------
    __cfg : dict
        Dictionary containing the pyGeo configuration
    __dvGeo : pyGeo.DVGeometry object
        geometry
    __dvCon : pyGeo.DVGConstraints object
        constraints
    """
    def __init__(self, cfg):
        """Instantiate the builder and save the configuration

        Parameters
        ----------
        cfg : dict
            Dictionary containing the pyGeo configuration

                geoFile : Name of file containing the FFD control points (.xyz) or the VSP geometry (.vsp3)
                vspOpts : Dictionary containing the VSP options
                disciplines : List of discipline names which points will be parametrized
                refAxis : Dictionary to create a reference axis. Each axis must be defined by a name and a dict of options (see DVGeometry)
                globalDV : Dictionary to create a global design variable. Each variable must be defined by a name and a dict of options (see DVGeometry)
                localDV : Dictionary to create a set of local design variables. Each set must be defined by a name and a dict of options (see DVGeometry)
                thckCon1D : Dictionary to create a thickness constraint along a line. Each constraint must be defined by a name and a dict of options (see DVConstraints)
                thckCon2D : Dictionary to create a thickness constraint along a surface. Each constraint must be defined by a name and a dict of options (see DVConstraints)
                volCon : Dictionary to create a volume constraint. Each constraint must be defined by a name and a dict of options (see DVConstraints)
                leteCon : Dictionary to create a set of LE/TE constraints. Each set must be defined by a name and a dict of options (see DVConstraints)
                linConShape : Dictionary to create a set of linear constraints on shape design variables. Each set must be defined by a name and a dict of options (see DVConstraints)
                vspVar : Dictionary to create a VSP design variable. Each variable must be defined by an identifier and a dict of options (see DVGeometryVSP)
        """
        # Sanity checks
        import os.path
        self.__cfg = cfg
        _, self.__cfg['format'] = os.path.splitext(cfg['geoFile'])
        if self.__cfg['format'] == '.xyz' and 'vspOpts' in cfg:
            raise RuntimeError('PyGeoBuilder - VSP options file cannot be provided with FFD file!\n')
        if self.__cfg['format'] != '.xyz' and self.__cfg['format'] != '.vsp3':
            raise RuntimeError('PyGeoBuilder - unsupported file format (supported .xyz, vsp3)!\n')

    def initialize(self, comm):
        """Instantiate and initialize all the geometry components

        Parameters
        ----------
        comm: mpi4py.MPI.Comm object
            MPI communicator
        """
        # Initialize the geometry parametrization
        if self.__cfg['format'] == '.xyz':
            from pygeo import DVGeometry
            self.__dvGeo = DVGeometry(self.__cfg['geoFile'])
        else:
            from pygeo import DVGeometryVSP
            self.__dvGeo = DVGeometryVSP(self.__cfg['geoFile'], comm=self.comm, **self.__opts)
        # Initialize the constraints
        from pygeo import DVConstraints
        self.__dvCon = DVConstraints()
        self.__dvCon.setDVGeo(self.__dvGeo)

    def get_mesh_coordinate_subsystem(self, scenario_name=None):
        """Return OpenMDAO component that wraps pyGeo
        """
        return PyGeoGeom(dvGeo=self.__dvGeo, dvCon=self.__dvCon, config=self.__cfg)

class PyGeoBuilderOMF(OMFlutBuilder):
    """pyGeo builder for OMFlut

    Attributes
    ----------
    see PyGeoBuilder
    """
    def __init__(self, cfg):
        """Instantiate and initialize all the geometry components

        Parameters
        ----------
        see PyGeoBuilder
        """
        # Sanity checks
        import os.path
        self.__cfg = cfg
        _, self.__cfg['format'] = os.path.splitext(cfg['geoFile'])
        if self.__cfg['format'] == '.xyz' and 'vspOpts' in cfg:
            raise RuntimeError('PyGeoBuilder - VSP options file cannot be provided with FFD file!\n')
        if self.__cfg['format'] != '.xyz' and self.__cfg['format'] != '.vsp3':
            raise RuntimeError('PyGeoBuilder - unsupported file format (supported .xyz, vsp3)!\n')
        # Initialize the geometry parametrization
        if self.__cfg['format'] == '.xyz':
            from pygeo import DVGeometry
            self.__dvGeo = DVGeometry(self.__cfg['geoFile'])
        else:
            from pygeo import DVGeometryVSP
            self.__dvGeo = DVGeometryVSP(self.__cfg['geoFile'], comm=self.comm, **self.__opts)
        # Initialize the constraints
        from pygeo import DVConstraints
        self.__dvCon = DVConstraints()
        self.__dvCon.setDVGeo(self.__dvGeo)

    def get_mesh(self):
        """Return OpenMDAO component that wraps pyGeo
        """
        return PyGeoGeom(dvGeo=self.__dvGeo, dvCon=self.__dvCon, config=self.__cfg, use_old_io=True)
