from utils import get_path
from tacs import elements, constitutive, functions
from mphys import MPhysVariables
import numpy as np

def get(pbl='static'):
    # define elements
    def element_callback(dvNum, compID, compDescript, elemDescripts, specialDVs, **kwargs):
        # Thickness
        def define_thickness(i):
            if i > 20:
                return 3e-3
            elif i > 17:
                return 4e-3
            elif i > 14:
                return 5e-3
            elif i > 11:
                return 6e-3
            elif i > 8:
                return 7e-3
            elif i > 5:
                return 8e-3
            elif i > 2:
                return 9e-3
            else:
                return 1e-2
        comp = compDescript.split('/')
        num_prt = int(comp[1].split('.')[1])
        num_seg = int(comp[2].split('.')[1])
        if comp[0] == 'RIBS':
            thickness = 3e-3
        elif comp[0] == 'SPARS':
            thickness = define_thickness(num_seg)
        else:
            thickness = define_thickness(num_prt)
        min_thickness = 0.001
        max_thickness = 0.02
        # Material properties
        rho = 2780.0 * 20 if pbl == 'modal' else 2780.0 # density, kg/m^3 (hack so that baseline design flutters)
        E = 73.1e9 # elastic modulus, Pa
        nu = 0.33 # poisson's ratio
        ys = 324.0e6 # yield stress, Pa
        # Setup (isotropic) property and constitutive objects
        prop = constitutive.MaterialProperties(rho=rho, E=E, nu=nu, ys=ys)
        # Set one thickness dv for every component
        con = constitutive.IsoShellConstitutive(prop, t=thickness, tNum=dvNum, tlb=min_thickness, tub=max_thickness)
        # For each element type in this component, pass back the appropriate tacs element object
        transform = None
        elem = elements.Quad4Shell(transform, con)
        return [elem]
    # define static problem
    def problem_setup(scenario_name, fea_assembler, problem):
        # Mass
        if scenario_name == 'cruise':
            problem.addFunction('mass', functions.StructuralMass)
        # Von-Mises failure criterion
        problem.addFunction('ks_fail', functions.KSFailure, ksWeight=50.0, safetyFactor=1.5)
        # Inertial relief gravity load
        if scenario_name == 'cruise':
            g = 1.0 * np.array([0.0, 0.0, -9.81]) # 1.0g (m/s^2)
        elif scenario_name == 'maneuver':
            g = 2.5 * np.array([0.0, 0.0, -9.81]) # 2.5g (m/s^2)
        else:
            raise RuntimeError(f'tacs_setup - undefined scenario: "{scenario_name}"!\n')
        problem.addInertialLoad(g)
    # define constraints
    def constraint_setup(scenario_name, fea_assembler, constraints):
        if scenario_name == 'cruise':
            # Adjacency
            cons = fea_assembler.createAdjacencyConstraint('adjacency')
            cid = fea_assembler.selectCompIDs(include='SPAR.00')
            cons.addConstraint('le_spar', compIDs=cid)
            cid = fea_assembler.selectCompIDs(include='SPAR.01')
            cons.addConstraint('te_spar', compIDs=cid)
            cid = fea_assembler.selectCompIDs(include='U_SKIN')
            cons.addConstraint('up_skin', compIDs=cid)
            cid = fea_assembler.selectCompIDs(include='L_SKIN')
            cons.addConstraint('lw_skin', compIDs=cid)
            constraints.append(cons)

    if pbl == 'static':
        return {
            'element_callback': element_callback,
            'problem_setup': problem_setup,
            'constraint_setup': constraint_setup,
            'coupling_loads': [MPhysVariables.Structures.Loads.AERODYNAMIC],
            'mesh_file': get_path('wingbox-L1-Order2.bdf', 'setup')
            }
    elif pbl == 'modal':
        return {
            'element_callback': element_callback,
            'num_modes': 6,
            'sigma': 1.,
            'exclude_modes': [1, 4],
            'components_tag': ['U_SKIN'],
            'mesh_file': get_path('wingbox-L3-Order2.bdf', 'setup'),
            'write_solution': False
        }
    else:
        raise RuntimeError('tacs_setup - Problem type not defined!\n')
