from utils import get_path
from tacs import elements, constitutive, functions
from mphys import MPhysVariables
import numpy as np

def get(pbl='static'):
    # define elements
    def element_callback(dvNum, compID, compDescript, elemDescripts, specialDVs, **kwargs):
        # Material properties (T. Brooks, Aerostructural tradeoffs for tow-steered composite wings)
        e11 = 117.9e9 # Young's modulus in direction '1, Pa
        e22 = 9.7e9 # Young's modulus in direction 2, Pa
        g12 = 4.8e9 # transverse shear modulus in planes 1-2, 1-3 and 2-3, Pa
        nu12 = 0.35 # Poisson's ratio in plane 1-2
        rho = 1.55e3 * 10 if pbl == 'modal' else 1.55e3 # density, kg/m^3
        t1 = 1648e6 # tensile strength in direction 1,  Pa
        c1 = 1034e6 # compressive strength in 1 direction 1, Pa
        t2 = 64e6 # tensile strength in 2 direction 2, Pa
        c2 = 228e6 # compressive strength in 2 direction 2, Pa
        s12 = 71e6 # shear strength in plane 1-2, Pa
        # Laminate layup (J. Dillinger, Static Aeroelastic Optimization of Composite Wings with Variable Stiffness Laminates)
        # NB: panel lengths are approximate; they will be constrained in the optimization problem
        if 'SKIN' in compDescript:
            # 0 deg biased layup with stiffeners aligned in the direction along TE spar
            ply_angles = np.deg2rad(np.array([0.0, -45.0, 45.0, 90.0]))
            panel_ply_fractions = np.array([44.41, 22.2, 22.2, 11.19]) / 100.0
            ref_axis = np.array([0.34968083, 0.93686889, 0.0])
            panel_length = 0.65
        else:
            # +/- 45 deg biased layup with stiffeners aligned in the vertical direction
            ply_angles = np.deg2rad(np.array([0.0, -45.0, 45.0, 90.0]))
            panel_ply_fractions = np.array([10.0, 35.0, 35.0, 20.0]) / 100.0
            ref_axis = np.array([0.0, 0.0, 1.0])
            if 'RIB' in compDescript:
                panel_length = 0.38
            elif 'SPAR' in compDescript:
                panel_length = 0.36
            else:
                raise RuntimeError(f'tacs_setup: properties for component "{compDescript}" not defined!\n')
        stiffener_ply_fractions = np.array([44.41, 22.2, 22.2, 11.19]) / 100.0 # 0 deg biased layup
        # Design variables (0: panel length, 1: stiffener pitch, 2: panel thickness, 3: NOT USED panel ply fractions, 4: stiffener height, 5: stiffener thickness, 6: NOT USED stiffener ply fractions)
        from setup.init_dv import dv_struct
        panel_length = dv_struct[dvNum]
        stiffener_pitch = dv_struct[dvNum + 1] #0.2 # m
        panel_thickness = dv_struct[dvNum + 2] #0.02 # m
        stiffener_heigth = dv_struct[dvNum + 3] #0.05 # m
        stiffener_thickness = dv_struct[dvNum + 4] #0.02 # m
        dv_scales = [1.0, 1.0, 100.0, 10.0, 100.0]
        # Create material property and othrotropic stiffened shell constitutive objects
        prop = constitutive.MaterialProperties(rho=rho,
            E1=e11, E2=e22, G12=g12, G13=g12, G23=g12,
            nu12=nu12,
            T1=t1, C1=c1, T2=t2, C2=c2, S12=s12)
        ply = constitutive.OrthotropicPly(1.0, prop)
        con = constitutive.BladeStiffenedShellConstitutive(panelPly=ply, stiffenerPly=ply,
            panelLength=panel_length, stiffenerPitch=stiffener_pitch, panelThick=panel_thickness, panelPlyAngles=ply_angles, panelPlyFracs=panel_ply_fractions, stiffenerHeight=stiffener_heigth, stiffenerThick=stiffener_thickness, stiffenerPlyAngles=ply_angles, stiffenerPlyFracs=stiffener_ply_fractions,
            panelLengthNum=dvNum, stiffenerPitchNum=dvNum+1, panelThickNum=dvNum+2, stiffenerHeightNum=dvNum+3, stiffenerThickNum=dvNum+4)
        con.setStiffenerPitchBounds(0.05, 0.5)
        con.setPanelThicknessBounds(0.002, 0.1)
        con.setStiffenerHeightBounds(0.002, 0.1)
        con.setStiffenerThicknessBounds(0.002, 0.1)
        # Create the element
        transform = elements.ShellRefAxisTransform(ref_axis)
        elem = elements.Quad4Shell(transform, con)
        return elem, dv_scales
    # define static problem
    def problem_setup(scenario_name, fea_assembler, problem):
        # Mass
        if scenario_name == 'cruise':
            problem.addFunction('mass', functions.StructuralMass)
        # Von-Mises failure criterion
        problem.addFunction('ks_fail', functions.KSFailure, ksWeight=100.0, safetyFactor=1.5)
        # Inertial relief gravity load
        if scenario_name == 'cruise':
            g = 1.0 * np.array([0.0, 0.0, -9.81]) # 1.0g (m/s^2)
        elif scenario_name == 'maneuver':
            g = 2.5 * np.array([0.0, 0.0, -9.81]) # 2.5g (m/s^2)
        else:
            raise RuntimeError(f'tacs_setup - undefined scenario: "{scenario_name}"!\n')
        problem.addInertialLoad(g)
    # define constraints
    def constraint_setup(scenario_name, fea_assembler, constraints):
        if scenario_name == 'cruise':
            # Adjacency
            adj = fea_assembler.createAdjacencyConstraint('adjacency_con')
            for g_id, g_name in {'SPAR.00': 'le_spar', 'SPAR.01': 'te_spar', 'U_SKIN': 'up_skin', 'L_SKIN': 'lw_skin'}.items():
                cids = fea_assembler.selectCompIDs(include=g_id)
                adj.addConstraint(conName=g_name+'_stiff_pitch', compIDs=cids, dvIndex=1)
                adj.addConstraint(conName=g_name+'_panel_thick', compIDs=cids, dvIndex=2)
                adj.addConstraint(conName=g_name+'_stiff_heigth', compIDs=cids, dvIndex=3)
                adj.addConstraint(conName=g_name+'_stiff_thick', compIDs=cids, dvIndex=4)
            constraints.append(adj)
            # Constraints between DV
            dv = fea_assembler.createDVConstraint('dv_con')
            dv.addConstraint(conName='panel_stiff_thick', dvIndices=[2, 4], dvWeights=[1.0, -1.0]) # adjacency panel/stiffener
            dv.addConstraint(conName='stiff_ar_max', dvIndices=[3, 4], dvWeights=[1.0, -10.0]) # stiffener max. aspect ratio
            dv.addConstraint(conName='stiff_ar_min', dvIndices=[3, 4], dvWeights=[1.0, -2.0]) # stiffener min. aspect ratio
            dv.addConstraint(conName='stiff_hgt_pitch', dvIndices=[3, 1], dvWeights=[2.0, -1.0]) # stiffener heigth/pitch
            constraints.append(dv)
            # Panel length
            lgt = fea_assembler.createPanelLengthConstraint('panel_con')
            lgt.addConstraint(conName='length', dvIndex=0)
            constraints.append(lgt)

    if pbl == 'static':
        return {
            'element_callback': element_callback,
            'problem_setup': problem_setup,
            'constraint_setup': constraint_setup,
            'coupling_loads': [MPhysVariables.Structures.Loads.AERODYNAMIC],
            'mesh_file': get_path('wingbox-L1-Order2.bdf', 'setup')
            }
    elif pbl == 'modal':
        return {
            'element_callback': element_callback,
            'num_modes': 6,
            'sigma': 1.,
            'exclude_modes': [1, 4],
            'components_tag': ['U_SKIN'],
            'mesh_file': get_path('wingbox-L3-Order2.bdf', 'setup'),
            'write_solution': False
        }
    else:
        raise RuntimeError('tacs_setup - Problem type not defined!\n')
